"""
Copyright (c) 2011 Socrata.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import re
from poster.encode import multipart_encode
from poster.streaminghttp import register_openers
from httplib2 import Http
from urllib import urlencode
from urllib2 import Request, urlopen

class SocrataBase:
    """Base class for all Socrata API objects"""

    def __init__(self, configuration):
        """
        Initializes a new Socrata API object with configuration
        options specified in standard ConfigParser format
        """

        print "SocrataBase init"

        self.config = configuration
        self.username, password, host = (self.config.get('credentials', 'user'),
            self.config.get('credentials', 'password'),
            self.config.get('server', 'host'))

        self.app_token  = self.config.get('credentials', 'app_token')

        self.api        = Http()
        self.url        = host
        self.id_pattern = re.compile('^[0-9a-z]{4}-[0-9a-z]{4}$')
        self.api.add_credentials(self.username, password)

        # For multipart upload/streaming
        register_openers()

    def _request(self, service, type, data = {}):
        """Generic HTTP request, encoding data as JSON and decoding the response"""

        print service, data

        response, content = self.api.request(
            self.url + service, type,
            headers = { 'Content-type:': 'application/json',
              'X-App-Token': self.app_token },
            body = json.dumps(data) )
        if content != None and len(content) > 0:
            response_parsed = json.loads(content)
            if hasattr(response_parsed, 'has_key') and \
                response_parsed.has_key('error') and response_parsed['error'] == True:
                print "Error: %s" % response_parsed['message']
            return response_parsed
        return None

    def _batch(self, data):
        payload = {'requests': data}
        return self._request('/batches', 'POST', payload)


class Dataset(SocrataBase):
    """Represents a Socrata Dataset, can be used for CRUD and more"""

    # Creates a new column, POSTing the request immediately
    def add_column(self, name, description='', type='text',
        hidden=False, rich=False, width=100):
        if not self.attached():
            return False
        data = { 'name': name, 'dataTypeName': type,
            'description': description, 'hidden': hidden,
            'width': width }
        if rich:
            data['format'] = {'formatting_option': 'Rich'}
        self.response = self._request("/views/%s/columns.json" % self.id,
            'POST', data)
        return self.response

    # Adds a new row by specifying an array of cells
    def add_row(self, data):
        if not self.attached():
            return False
        self.response = self._request("/views/%s/rows.json" % self.id,
            'POST', data)

    # For batch row importing, returns a dict to be POST'd
    # Takes same array of cells as add_row
    def add_row_delayed(self, data):
        if not self.attached():
            return False
        return {'url': "/views/%s/rows.json" % self.id,
                'requestType': 'POST',
                'body': json.dumps(data)}

    # Is this class currently associated with an existing dataset?
    def attached(self):
        return self.is_id(self.id)

    # Creates a new dataset by sending a POST request, saves four-four ID
    def create(self, name, description = '', tags = [], public = True):
        self.error = False
        data = { 'name': name, 'description': description }
        if public:
            data['flags'] = ['dataPublicRead']
        if tags.count > 0:
            data['tags'] = tags
        
        response = self._request('/views.json', 'POST', data)
        if response.has_key('error'):
            self.error = response['message']
            if response['code'] == 'authentication_required':
                raise RuntimeError('You must specify proper authentication credentials')
            elif response['code'] == 'invalid_request':
                raise DuplicateDatasetError(name)
            else:
                raise RuntimeError('API Error ' + response['message'])
        self.id = response['id']
        return True

    # Deletes the active dataset
    def delete(self):
        self.error = False
        response = self._request("/views.json?id=%s&method=delete" % self.id,
            'DELETE', None)
        return response == None

    # Call the search service with optional params
    def find_datasets(self, params={}, page=1, results=[]):
        self.error = False
        print "page is {0}".format(page)
        params['page'] = page
        sets = self._request("/api/search/views.json?%s" % urlencode(params), "GET")
        if 'results' in sets and len(sets['results']) > 0:
            results += sets['results']
            print 'sets count is {0}'.format(len(sets['results']))
            print 'results count is {0}'.format(len(results))
            page = page + 1
            self.find_datasets(params, page, results)
        return results

    def find_view_detail(self, table_id):
        self.error = False
        view_detail = self._request('/api/views/%s/rows.json' % table_id, "GET")
        return view_detail 

    def metadata(self):
        self.error = False
        self.response = self._request("/views/%s.json" % self.id, 'GET')
        return self.response['metadata']

    def attachments(self):
        metadata = self.metadata()
        if metadata == None or (not metadata.has_key('attachments')):
            return []
        return metadata['attachments']

    def attach_file(self, filename):
        metadata = self.metadata()
        if not metadata.has_key('attachments'):
            metadata['attachments'] = []

        response = self.multipart_post('/assets', filename)
        if not response.has_key('id'):
            print "Error uploading file to assets service: no ID returned: %s" % response
            return
        attachment = {'blobId': response['id'],
            'name': response['nameForOutput'],
            'filename': response['nameForOutput']}
        metadata['attachments'].append(attachment)
        self._request("/views/%s.json" % self.id, 'PUT', {'metadata':metadata})

    def multipart_post(self, url, filename, field='file'):
        datagen, headers       = multipart_encode({field: open(filename, "rb")})
        headers['X-App-Token'] = self.app_token

        request = Request("%s%s" % (self.url, url), datagen, headers)
        response = urlopen(request).read()
        return json.loads(response)

    # Is the string 'id' a valid four-four ID?
    def is_id(self, id):
        return self.id_pattern.match(id) != None

    def use_existing(self, id):
        if self.is_id(id):
            self.id = id
        else:
            return False

    def short_url(self):
        return self.config.get('server', 'host') + "/d/" + str(self.id)


class DuplicateDatasetError(Exception):
    """Raised if a dataset with the specified name already exists"""
    def __init__(self, name):
        self.name = name
    def __str__(self):
        return "Duplicate dataset with name '%s'" % self.name



import ConfigParser
from socrata_python.Socrata import *

'''
WindyPie is a Python module that allows you to easily interact with
ALL the Socrata Data Portals

Copyright (C) 2012, Code for America
This is open source software, released under a standard 2-clause
BSD-style license; see the file LICENSE for details.
'''

class WindyPie(object):
    '''Wraps all interaction with Socrata API'''
    def __init__(self, url=None, socrata_adapter=None):
        if not socrata_adapter:
            socrata_adapter = SocrataPythonAdapter(url)
        self._views = WindyPie.Views(socrata_adapter)
        self._version = '0.0.3'

    @property
    def views(self):
        '''list of views'''
        return self._views

    @property
    def version(self):
        '''return version number of this software'''
        return self._version

    class View(object):
        '''A model object that represents a socrata view'''
        def __init__(self, data):
            self._data = data
            self._field_names = []
            self._collection = self.__init_collection()

        def __init_collection(self):
            '''load raw socrata data into more managable collections of columns and rows'''
            # socrata view meta data is lengthy! we load only the filed (aka column) string
            # values into our fields property
            for column in self.columns:
                self._field_names.append(str(column['fieldName']))
            # socrata rows include only the data, this creates a list of dictionary objects
            # (keyed by field name) so that rows can be output in an easier to manage 
            # format such as json
            document_collection = []
            for row in self.rows:
                document = DotDict() # using DotDict class for dot notation of row fields
                for i in range(0,len(row)):
                    document[self.fields[i]] = str(row[i])
                document_collection.append(document)
            return document_collection

        @property
        def fields(self):
            '''all field name strings from columns'''
            return self._field_names

        @property
        def rows(self):
            '''the list of raw, socrata row objects for the view'''
            return self._data['data']

        @property
        def columns(self):
            '''the list of raw, socrata column objects for the view'''
            return self._data['meta']['view']['columns']

        @property
        def collection(self):
            '''the list of row objects for the view as python dicts'''
            return self._collection

        def __search_for_rows(self, field, value):
            '''find subset of rows with fields equal to value'''
            collection = []
            for row in self.collection:
                if value == row[str(field)]:
                    collection.append(row)
            return collection

        def __getattr__(self, name):
            '''handle requests to search for rows by field name dynamically'''
            # if the request is not in the ballpark then reject it
            if not name.startswith('collection_by_'):
                raise AttributeError
            def method(*args):
                field = name[len('collection_by_'):]
                return self.__search_for_rows(field, args[0])
            return method

    class Views(object):
        '''Represents collection of Socrata views'''
        def __init__(self, adapter):
            self._adapter = adapter

        def __call__(self, view_id=None, view_name=None):
            '''Get views from Socrata'''
            if None == view_id:
                return self.query(view_name)
            else:
                return WindyPie.View(self._adapter.find_by_id(view_id))

        def query(self, view_name):
            '''query socrata views table and return based on paramter filter values'''
            params = {}
            if None != view_name:
                params['name'] = view_name
            return self._adapter.query_views(params) 

class SocrataPythonAdapter:
    '''Manage working with Socrata's odd python library until it is rewritten'''
    def __init__(self, url, user='', password='', token=''):
        self.url = url
        self.user = user
        self.password = password 
        self.token = token
        self.set_configuration()
        self.dataset = Dataset(self.config)

    def find_by_id(self, view_id):
        '''query socrata instance to return view with passed in id'''
        return self.dataset.find_view_detail(view_id)

    def query_views(self, params={}):
        '''polls through all available views in the socrata db and returns them in a list'''
        #XXX this is a blocking call, polling 200 at a time could take time to complete!
        params['limit'] = 200
        return self.dataset.find_datasets(params, 1)

    def set_configuration(self):
        '''socrata python library uses ConfigParser, so must we'''
        self.config = ConfigParser.ConfigParser()
        self.config.add_section('credentials')
        self.config.add_section('server')
        self.config.set('credentials', 'app_token', self.token)
        self.config.set('credentials', 'user', self.user)
        self.config.set('credentials', 'password', self.password)
        self.config.set('server', 'host', self.url)

class DotDict(dict):
    '''
    dict that supports dot notation for elements
    this is a cherry pick of more comprehensive solutions
    see: https://github.com/vkuznet/DotDict
    '''
    def __getattr__(self, attr):
        return self.get(attr, None)
    __setattr__= dict.__setitem__
    __delattr__= dict.__delitem__



__author__="Aiden Scandella"
__date__ ="$Jan 11, 2010 5:06:00 PM$"

from setuptools import setup,find_packages

setup (
  name = 'Socrata-Python',
  version = '0.2',
  packages = find_packages(),

  install_requires=['feedparser','httplib2','poster'],

  author = 'Aiden Scandella',
  author_email = 'aiden.scandella@socrata.com',

  url = 'http://www.socrata.com',
  license = '',
  long_description= 'A native Python implementation of Socrata\'s REST API, using JSON'
)



#!/usr/bin/python

import Socrata
import ConfigParser
import sys
import urllib
import time
import feedparser
from xml.dom import minidom

REDDIT_RSS = "http://www.reddit.com/.rss"



def create_dataset_with_columns(dataset, title = 'RSS Feed Dataset', description = ''):
    """Creates a new Socrata dataset with columns for an RSS feed"""
    try:
        dataset.create(title, description)
    except Socrata.DuplicateDatasetError:
        print "This dataset already exists."
        return False

    dataset.add_column('Title', '', 'text', False, False, 300)
    dataset.add_column('URL', '', 'url', False, False, 300)
    dataset.add_column('Date', '', 'date')
    
    return

if __name__ == "__main__":
    # Default to Reddit for an example
    feed_url = REDDIT_RSS
    # Else take their feed from command line args
    if len(sys.argv) > 1:
        feed_url = sys.argv[1]
    print "Downloading RSS feed from " + str(feed_url)
    
    rss = feedparser.parse(feed_url)
    
    cfg = ConfigParser.ConfigParser()
    cfg.read('socrata.cfg')
    dataset = Socrata.Dataset(cfg)

    print "Searching for existing dataset"
    existing = dataset.find_datasets({'q':'RSS Feed Dataset',
        'for_user': dataset.username})[0]
    if existing['count'] > 0:
        print "Dataset exists, using it"
        dataset.use_existing(existing['results'][0]['id'])
    else:
        print "Creating dataset in Socrata"
        create_dataset_with_columns(dataset)

    if  dataset:
        batch_requests = []
        # Extract relevant information
        for item in rss.entries:
            data          = {}
            data['Title'] = item.title
            data['URL']   = item.link
            data['Date']  = time.strftime("%m/%d/%Y %H:%M:%S", item.date_parsed)
            batch_requests.append(dataset.add_row_delayed(data))


        dataset._batch(batch_requests)
        print "You can now view the dataset:"
        print dataset.short_url()
    else:
        print "There was an error creating your dataset."
    
    print "\nFinished"



'''
Setup and installation for the package.
'''

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name="WindyPie",
    version="0.0.3",
    url="https://github.com/codeforamerica/windypie",
    author="Jesse Bounds",
    author_email="jesse@codeforamerica.org",
    description="An easy-to-use wrapper for the Socrata API",
    packages=[
        'windypie', 'socrata_python'
    ],
    install_requires=[
        'httplib2',
        'configparser',
        'ordereddict',
        'poster'
    ],
    license='BSD-style license; see the file LICENSE for details.',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
)
